#ifndef __DUNE_MMDG_MMDGPROBLEM3_HH__
#define __DUNE_MMDG_MMDGPROBLEM3_HH__

#include <cmath>
#include <dune/mmdg/problems/coupleddgproblem.hh>

//a coupled dg problem for dim = 2,
//taken from [Antonietti et al. (2019): Example 6.1]
template<class Coordinate, class Scalar = double>
class MMDGProblem3 : public CoupledDGProblem<Coordinate, Scalar>
{
public:
  static constexpr int dim = Coordinate::dimension;
  using Base = CoupledDGProblem<Coordinate, Scalar>;
  using Matrix = typename Base::Matrix;

  //constructor
  MMDGProblem3 (const Scalar d) : d_(d) {}

  //the exact bulk solution at position pos
  Scalar exactSolution (const Coordinate& pos) const
  {
    Scalar xPlusy = pos[0] + pos[1];// pos * Coordinate(1.0);
    Scalar solution = std::exp(xPlusy);
    return (xPlusy < 1.0) ? solution :
      0.5 * solution + std::exp(1.0) * (0.5 + 1.5 / sqrt(2.0) * d_);
  }

  //the exact solution on the interface at position pos
  Scalar exactInterfaceSolution (const Coordinate& pos) const
  {
    return std::exp(1.0) * ( 1.0 + 1.0 / sqrt(2.0) * d_ );
  }

  //indicates whether an exact solution is implemented for the problem
  bool hasExactSolution () const
  {
    return true;
  };

  //bulk source term at position pos
  Scalar q (const Coordinate& pos) const
  {
    Scalar xPlusy = pos[0] + pos[1]; //pos * Coordinate(1.0);
    Scalar source = -std::exp(xPlusy);
    return (xPlusy < 1.0) ? 2.0 * source : source;
  }

  //interface source term at position pos
  Scalar qInterface (const Coordinate& pos) const
  {
    constexpr Scalar iSource = std::exp(1.0) / sqrt(2.0);
    return iSource;
  }

  //aperture d of the fracture at position pos
  Scalar aperture (const Coordinate& pos) const
  {
    return d_;
  }

  //permeability per aperture of the fracture in normal direction at position pos
  Scalar Kperp (const Coordinate& pos) const
  {
    return 1.0 / d_;
  }

  //returns the recommended quadrature order to compute an integral
  //over x * q(x)
  int quadratureOrder_q () const
  {
    return 10;
  }

  //returns the recommended quadrature order to compute an integral
  //over x * boundary(x)
  int quadratureOrderBoundary () const
  {
    return 10;
  }

private:
  Scalar d_; //aperture
};

#endif
